﻿using UnityEngine;
using System;
using System.Collections;
using System.Collections.Generic;

namespace AT
{
	namespace UBF
	{
		namespace Behaviors
		{
			/// <summary>
			/// Detects when it is close to a teamate and attempts to steer away
			/// </summary>
			class DecentAgent_AvoidTeamates : AT.UBF.Behavior
			{
				/// <summary>
				/// Priority when the two tanks are within shorter buffer of each other
				/// </summary>
				public float higherPriority = 1.0f;

				/// <summary>
				/// Minimum priority when the two tanks are within longerBuffer of each other. Priority increases linearly to higherPriority
				/// </summary>
				public float lowerPriority = 0f;

				/// <summary>
				/// Distance inside which higherPriority is used
				/// </summary>
				public float shorterBuffer = 4.0f;

				/// <summary>
				/// Distance inside which at least lowerPriority is used
				/// </summary>
				public float longerBuffer = 6.0f;

				/// <summary>
				/// How fast to turn when inside longer buffer
				/// </summary>
				public float turnRightMultiplier = 0.5f;

				/// <summary>
				/// Draw lines in scene mode to describe throught process of the behavior
				/// </summary>
				public bool debuggingLines = false;

				/// <summary>
				/// Generate an action dictated by the logic of this behavior
				/// </summary>
				/// <param name="state">State of the environment as seen by the tank</param>
				/// <returns>Action corresponding to the precepts for the behavior</returns>
				protected override AT.UBF.Action genActionInternal(AT.UBF.State state)
				{
					AT.UBF.Action action = new AT.UBF.Action();

					var mypos = state.self.getPosition();

					List<string> originalMessages;
					var cstate = DecentAgentUtil.GetCombinedStateFromMessages(state, out originalMessages);

					//Get a list of teamates
					var teamates = cstate.teamates;


					//Check if we are on the inside of the corner
					if (teamates != null && teamates.Count > 0)
					{
						//Get the closest teamate
						teamates.Sort((x, y) => ((x.position - mypos).magnitude).CompareTo((y.position - mypos).magnitude));

						var teamate = teamates[0];

						if (debuggingLines)
						{
							Debug.DrawLine(mypos, teamate.position, Color.black);
							Debug.DrawLine(mypos, mypos + (teamate.position - mypos).normalized * longerBuffer, Color.blue);
						}

						if ((teamate.position - mypos).magnitude <= longerBuffer)
						{
							//Find and set the priority of this action
							float priority = (higherPriority - lowerPriority) / (longerBuffer - shorterBuffer) * (mypos - teamate.position).magnitude;
							action.SetPriority(priority, voteSourceID);

							float maxWallLength = DecentAgentUtil.getMaxWallSize(state.walls);
							var corners = DecentAgentUtil.findCornersTankIsInsideOf(state.walls, mypos, maxWallLength, false);
							//Get the closest corner
							corners.Sort((x, y) => ((x.first - mypos).magnitude).CompareTo((y.first - mypos).magnitude));
							var corner = corners[0];
							//Get the closest outside endpoint of the corner walls
							var wallEnd = (corner.second.second - mypos).normalized.magnitude < (corner.third.second - mypos).normalized.magnitude ? corner.second.second : corner.third.second;

							//Decide which way to turn
							var angleToTeamate = Util.getRelativeAngleToTarget(mypos, teamate.position);
							var angleTracks = state.self.getTracksRotation();

							if (angleToTeamate > 270 && angleTracks < 90)
							{
								angleToTeamate -= 360;
							}
							if (angleTracks > 270 && angleToTeamate < 90)
							{
								angleTracks -= 360;
							}


							float turnRight = angleTracks < angleToTeamate ? 1f : -1f;

							if (debuggingLines)
							{
								Debug.DrawLine(mypos, teamate.position, Color.red);
							}

							action.SetTurnTracksRight(turnRight * turnRightMultiplier, voteSourceID);
						}
					}

					return action;
				}

				/// <summary>
				/// Write the values to the dictionary
				/// </summary>
				/// <param name="dict"></param>
				/// <param name="saveLastAction"></param>
				public override void _ToDict(Dictionary<string, object> dict, bool saveLastAction)
				{
					dict["HigherPriority"] = higherPriority;
					dict["ShorterBuffer"] = shorterBuffer;
					dict["LowerPriority"] = lowerPriority;
					dict["LongerBuffer"] = longerBuffer;
					dict["TurnRightMultiplier"] = turnRightMultiplier;
					dict["DebuggingLines"] = debuggingLines;

				}

				/// <summary>
				/// Get the class values from the dictionary
				/// </summary>
				/// <param name="dict">Dictionary to restore from</param>
				public override void _FromDict(Dictionary<string, object> dict)
				{
					string bName = this.GetType().Name;
					higherPriority = Util.FloatFromJsonDict("HigherPriority", higherPriority, bName, dict);
					shorterBuffer = Util.FloatFromJsonDict("ShorterBuffer", shorterBuffer, bName, dict);
					lowerPriority = Util.FloatFromJsonDict("LowerPriority", lowerPriority, bName, dict);
					longerBuffer = Util.FloatFromJsonDict("LongerBuffer", longerBuffer, bName, dict);
					turnRightMultiplier = Util.FloatFromJsonDict("TurnRightMultiplier", turnRightMultiplier, bName, dict);
					debuggingLines = false;
					try
					{
						debuggingLines = (bool)dict["DebuggingLines"];
					}
					catch (KeyNotFoundException)
					{
						Debug.Log(String.Format("'{0}' key not found in {1} Arguments. Using default of {2}.", "DebuggingLines", debuggingLines, false));
					}
					catch (InvalidCastException)
					{
						Debug.LogError(String.Format("'{0}' value in {1} Arguments is not a valid float", "DebuggingLines", debuggingLines));
					}
				}
			}
		}
	}
}
