﻿using UnityEngine;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;


namespace AT
{
	namespace UBF
	{
		namespace Behaviors
		{
			/// <summary>
			/// This behavior attempts to dodge the closest incoming bullet. 
			/// Note:It votes on the driveForward part of the action
			/// </summary>
			public class DecentAgent_DodgeClosestIncomingBullet : AT.UBF.Behavior
			{
				/// <summary>
				/// Priority of the behavior if there is an incoming bullet
				/// </summary>
				public float priority = 0.8f;

				/// <summary>
				/// Buffer inside which to turn the tank opposite of where it is going in order to not run into wall
				/// </summary>
				public float shortWallBuffer = 3f;

				/// <summary>
				/// Buffer outside of which to stop turning the tank opposite of where it is going in order to not run into wall
				/// </summary>
				public float longWallBuffer = 6f;

				/// <summary>
				/// Should reverse course to not hit wall?
				/// </summary>
				private bool reverse = false;

				/// <summary>
				/// Shows the thought process of this behavior in the scenes window if enabled
				/// </summary>
				public bool debuggingLines = false;

				/// <summary>
				/// Generate an action to dodge the closest incoming bullet
				/// </summary>
				/// <param name="state">State of the environment as seen by the tank</param>
				/// <returns>Action comensurate with the given precepts of the tank</returns>
				protected override AT.UBF.Action genActionInternal(AT.UBF.State state)
				{
					AT.UBF.Action action = new AT.UBF.Action();

					var mypos = state.self.getPosition();
					var myvel = state.self.getVelocity();

					List<string> originalMessages;
					var cstate = DecentAgentUtil.GetCombinedStateFromMessages(state, out originalMessages);

					//If there are visible bullets, then make a action to vote with
					if (cstate.combinedVisibleBullets.Count > 0)
					{
						//Get the dimensions of the tank
						Vector2 tankBufferLength = state.bodyLength.first - state.bodyLength.second;
						Vector2 tankBufferWidth = state.bodyWidth.first - state.bodyWidth.second;

						//Multiplier to use to make sure the tank goes the way you actually want it too
						float goingForwardMultiplier = 1.0f;

						//Get the closest incoming bullet
						Tuple<Vector2, Vector2> bullet = DecentAgentUtil.getClosestIncomingBullet(mypos, cstate.combinedVisibleBullets);

						if (bullet != null)
						{
							//Get the projection point onto the line going through the centerline of the tank
							//Note this is only used in the debug drawing of lines.
							Vector2 projPoint = DecentAgentUtil.projectVectortoLine(new Tuple<Vector2, Vector2>(mypos, myvel), bullet.first);

							//Get the possible hit point of the bullet with the tank.
							//Note that this point is the point that the bullet will intersect the centerline of the tank assuming that the tank drives stright there with a turnTracksRight =0
							Vector2 impactPoint = DecentAgentUtil.getIntersectionPoint(new Tuple<Vector2, Vector2>(mypos, myvel), bullet);

							//Get the realtive angle to the point of impact to figure out if it is in front or behind the tank
							float angleToImpact = Util.getRelativeAngleToTarget(mypos, impactPoint);
							if (Math.Abs(state.self.getTracksRotation() - angleToImpact) < 90f)
							{
								goingForwardMultiplier = -1.0f;
							}
							else
							{
								goingForwardMultiplier = 1f;
							}

							var wall = DecentAgentUtil.getClosestWall(mypos, state.walls);
							var distToWall = Util.distToLineSegment(mypos, wall.first, wall.second);
							var angleToWall = Util.getRelativeAngleToTarget(mypos, distToWall.second);

							//There are problems when getting close to walls. Try not to run into one
							if (distToWall.first < longWallBuffer)
							{
								if (distToWall.first < shortWallBuffer)
								{
									reverse = true;
								}
							}
							else
							{
								reverse = false;
							}

							if (reverse)
							{
								goingForwardMultiplier *= -1;
							}

							//Draw the lines that show what this behavior is actually thinking
							if (debuggingLines)
							{
								Debug.DrawLine(mypos - 3 * myvel, mypos + 3 * myvel, Color.magenta);
								Debug.DrawLine(projPoint, mypos, Color.grey);
								Debug.DrawLine(projPoint, bullet.first, Color.grey);
								Debug.DrawLine(bullet.first, mypos, Color.black);
								Debug.DrawLine(bullet.first, impactPoint, Color.red);
								Debug.DrawLine(impactPoint, mypos - tankBufferLength / 2, Color.cyan);
								Debug.DrawLine(impactPoint, mypos + tankBufferLength / 2, Color.cyan);
								Debug.DrawRay(state.self.getPosition(), new Vector2((float)Math.Cos(state.self.getTurretRotation().DegreesToRadians()), (float)Math.Sin(state.self.getTurretRotation().DegreesToRadians())), Color.white);
								Debug.Log("Driveforward: " + goingForwardMultiplier);

								//Draw lines that mark the edges of the tank
								//Debug.DrawLine(state.self.getPosition() - tankBufferLength / 2 - tankBufferWidth / 2, state.self.getPosition() + tankBufferLength / 2 - tankBufferWidth / 2, Color.cyan);
								//Debug.DrawLine(state.self.getPosition() - tankBufferLength / 2 - tankBufferWidth / 2, state.self.getPosition() - tankBufferLength / 2 + tankBufferWidth / 2, Color.cyan);
								//Debug.DrawLine(state.self.getPosition() + tankBufferLength / 2 + tankBufferWidth / 2, state.self.getPosition() + tankBufferLength / 2 - tankBufferWidth / 2, Color.cyan);
								//Debug.DrawLine(state.self.getPosition() + tankBufferLength / 2 + tankBufferWidth / 2, state.self.getPosition() - tankBufferLength / 2 + tankBufferWidth / 2, Color.cyan);
							}

							//Set the action
							action.SetDriveForward(0.9f * goingForwardMultiplier, voteSourceID);
							action.SetPriority(priority, voteSourceID);
						}
					}

					return action;
				}

				/// <summary>
				/// This was copied from the original DodgeBullet method 
				/// </summary>
				/// <param name="bulletLine"></param>
				/// <param name="bodyLength"></param>
				/// <param name="tankPos"></param>
				/// <returns></returns>
				private Tuple<Vector2, Vector2> getMaxTankWidthLine(Tuple<Vector2, Vector2> bulletLine, Tuple<Vector2, Vector2> bodyLength, Vector2 tankPos)
				{
					Vector2 v = bulletLine.second;
					Vector2 numerator = new Vector2(-v.y, v.x);

					Vector2 tankSize = (bodyLength.first - bodyLength.second) * 1.2f;
					float scaling = Mathf.Sqrt(Mathf.Pow(tankSize.x, 2) + Mathf.Pow(tankSize.y, 2)) / 2f;

					Vector2 p3 = tankPos + numerator / Mathf.Sqrt(Mathf.Pow(v.x, 2) + Mathf.Pow(v.y, 2)) * scaling;
					Vector2 p4 = tankPos + numerator / Mathf.Sqrt(Mathf.Pow(v.x, 2) + Mathf.Pow(v.y, 2)) * -scaling;

					return new Tuple<Vector2, Vector2>(p3, p4);
				}

				/// <summary>
				/// Write the values to the dictionary
				/// </summary>
				/// <param name="dict"></param>
				/// <param name="saveLastAction"></param>
				public override void _ToDict(Dictionary<string, object> dict, bool saveLastAction)
				{
					dict["Priority"] = priority;
					dict["DebuggingLines"] = debuggingLines;
					dict["ShortWallBuffer"] = shortWallBuffer;
					dict["LongWallBuffer"] = longWallBuffer;
					dict["Reverse"] = reverse;
				}

				/// <summary>
				/// Get the class values from the dictionary
				/// </summary>
				/// <param name="dict">Dictionary to restore from</param>
				public override void _FromDict(Dictionary<string, object> dict)
				{
					string bName = this.GetType().Name;
					priority = Util.FloatFromJsonDict("Priority", priority, bName, dict);
					shortWallBuffer = Util.FloatFromJsonDict("ShortWallBuffer", shortWallBuffer, bName, dict);
					longWallBuffer = Util.FloatFromJsonDict("LongWallBuffer", longWallBuffer, bName, dict);
					debuggingLines = DecentAgentUtil.BoolFromJsonDict("DebuggingLines", debuggingLines, bName, dict);
					reverse = DecentAgentUtil.BoolFromJsonDict("Reverse", reverse, bName, dict);
				}
			}
		}
	}
}
