﻿using UnityEngine;
using System;
using System.Collections;
using System.Collections.Generic;

namespace AT
{
	namespace UBF
	{
		namespace Behaviors
		{
			/// <summary>
			/// Rotates the tank to be oriented perpendicular to the first tank in the current tanks visible list
			/// Note: if there are no visible enemy tanks, this does not vote
			/// </summary>
			public class DecentAgent_RotatePerpendicularToEnemy : AT.UBF.Behavior
			{
				/// <summary>
				/// Priority of the action when the enemy is in sight
				/// </summary>
				public float priorityWhenEnemyInSight = 0.7f;

				/// <summary>
				/// Priority when the enemy is shooting at you
				/// </summary>
				public float priorityWhenBeingShotAt = 0.2f;

				/// <summary>
				/// Buffer inside whcih to care that bullets are getting so close
				/// </summary>
				public float bulletBuffer = 4;

				/// <summary>
				/// Shows the thought process of this behavior in the scenes window if enabled
				/// </summary>
				public bool debuggingLines = false;

				/// <summary>
				/// Generate an action for this class
				/// </summary>
				/// <param name="state">State as seen by the current tank</param>
				/// <returns>Action to rotate perpendicular to the enemy</returns>
				protected override Action genActionInternal(State state)
				{
					var action = new AT.UBF.Action();

					var mypos = state.self.getPosition();

					//Get the combined state from the messages
					List<string> originalMessages;
					var cstate = DecentAgentUtil.GetCombinedStateFromMessages(state, out originalMessages);
					var bullets = DecentAgentUtil.getIncomingBullets(mypos, state.visibleBullets, bulletBuffer, false);

					//If there are no incoming bullets or visible enemy tanks, don't do anything
					if (cstate.combinedVisibleEnemyTanks.Count < 1 && bullets.Count < 1)
					{
						return action;
					}
					//If there is an enemy tank visible, then rotate perpendicular to that first
					else if (cstate.combinedVisibleEnemyTanks.Count > 0)
					{
						//Get the first visible tank
						var targetPos = DecentAgentUtil.getClosestEnemyTank(state, cstate.combinedVisibleEnemyTanks, false);

						//Get the trackRight value to use to get perpendicular to the first visible tank
						float trackRight = DecentAgentUtil.findShortestTrackRotationPerpendicularToTarget(state.self.getPosition(), targetPos.position, state.self.getTracksRotation(), false);

						//Used for debugging
						if (debuggingLines)
						{
							//Debug.Log("Track right setting: " + trackRight);
							Debug.DrawLine(state.self.getPosition(), targetPos.position, Color.green);
						}
						//Set the action
						action.SetTurnTracksRight(trackRight, voteSourceID);

						//If there is a visible enemy tank, use that priority, otherwise use the being shot at priority
						action.SetPriority(priorityWhenEnemyInSight, voteSourceID);
					}
					//Else there is an incoming bullet we need to get perpendicular to
					else
					{
						//Get the first visible tank
						var targetPos = DecentAgentUtil.getClosestIncomingBullet(mypos, bullets, bulletBuffer, false);

						//Get the trackRight value to use to get perpendicular to the first visible tank
						float trackRight = DecentAgentUtil.findShortestTrackRotationPerpendicularToTarget(mypos, targetPos.first, state.self.getTracksRotation(), false);

						//Used for debugging
						if (debuggingLines)
						{
							//Debug.Log("Track right setting: " + trackRight);
							Debug.DrawLine(state.self.getPosition(), targetPos.first, Color.green);
						}

						//Set the action
						action.SetTurnTracksRight(trackRight, voteSourceID);

						//If there is a visible enemy tank, use that priority, otherwise use the being shot at priority
						action.SetPriority(priorityWhenBeingShotAt, voteSourceID);
					}
					return action;
				}

				/// <summary>
				/// Write the priorityWhenEnemyInSight to the dictionary
				/// </summary>
				/// <param name="dict">Dictionary to write to</param>
				/// <param name="saveLastAction"></param>
				public override void _ToDict(Dictionary<string, object> dict, bool saveLastAction)
				{
					dict["PriorityWhenEnemyInSight"] = priorityWhenEnemyInSight;
					dict["PriorityWhenBeingShotAt"] = priorityWhenBeingShotAt;
					dict["BulletBuffer"] = bulletBuffer;
					dict["DebuggingLines"] = debuggingLines;
				}

				/// <summary>
				/// Get the priorityWhenEnemyInSight value from the dictionary
				/// </summary>
				/// <param name="dict">Dictionary to read from</param>
				public override void _FromDict(Dictionary<string, object> dict)
				{
					string bName = this.GetType().Name;
					priorityWhenEnemyInSight = Util.FloatFromJsonDict("PriorityWhenEnemyInSight", priorityWhenEnemyInSight, bName, dict);
					priorityWhenBeingShotAt = Util.FloatFromJsonDict("PriorityWhenBeingShotAt", priorityWhenBeingShotAt, bName, dict);
					bulletBuffer = Util.FloatFromJsonDict("BulletBuffer", bulletBuffer, bName, dict);
					debuggingLines = DecentAgentUtil.BoolFromJsonDict("DebuggingLines", debuggingLines, bName, dict);
				}
			}
		}
	}
}
