﻿using UnityEngine;
using System;
using System.Collections;
using System.Collections.Generic;

namespace AT
{
	namespace UBF
	{
		namespace Behaviors
		{
			/// <summary>
			/// Detects when it is close to a corner and attempts to steer away
			/// </summary>
			class DecentAgent_RotatePerpendicularToWalls : AT.UBF.Behavior
			{
				/// <summary>
				/// Preffered distance to the wall when following parallel to it
				/// </summary>
				public float prefferedDistanceToWall = 2.5f;

				/// <summary>
				/// Priority for the tank to get perpendicular to the wall when it is within some buffer of the wall
				/// </summary>
				public float perpendicularToWallPriority = 1f;

				/// <summary>
				/// Priority when the tank is within the short buffer of a wall or corner
				/// </summary>
				public float higherPriority = 1.0f;

				/// <summary>
				/// Priority when the tank is within the long buffer of a wall or corner. The actual priority increases linearly with decreasing distance until the tank is within shorterBuffer
				/// </summary>
				public float lowerPriority = 0f;

				/// <summary>
				/// Buffer from wall or corner in which the highest priority value is used all the time
				/// </summary>
				public float shorterBuffer = 4.0f;

				/// <summary>
				/// Specifies the distance from a wall or corner this action will start voting
				/// </summary>
				public float longerBuffer = 7.0f;

				/// <summary>
				/// Degree to which the max rotation speed of the tank is used (Between 0 and 1)
				/// </summary>
				public float turnRightMultiplier = 0.5f;

				/// <summary>
				/// Buffer inside which to get perpendicular to a corner wall
				/// </summary>
				public float wallBuffer = 6f;

				/// <summary>
				/// Buffer inside which to get perpendicular to any wall (not just corner walls)
				/// </summary>
				public float normalWallBuffer = 5;

				/// <summary>
				/// Shows debugging lines in the scene view
				/// </summary>
				public bool debuggingLines = false;

				/// <summary>
				/// Creates an action that votes to get the agent perpendicular to the closest wall. Does not vote if not within certain buffer distances of a wall
				/// </summary>
				/// <param name="state">State of the environment as seen by the tank</param>
				/// <returns>Action for the tank to hopefully complete</returns>
				protected override AT.UBF.Action genActionInternal(AT.UBF.State state)
				{
					AT.UBF.Action action = new AT.UBF.Action();

					Vector2 tankBufferLength = state.bodyLength.first - state.bodyLength.second;
					Vector2 tankBufferWidth = state.bodyWidth.first - state.bodyWidth.second;
					var internalBuffer = (tankBufferLength / 2 + tankBufferWidth / 2).magnitude;

					var mypos = state.self.getPosition();

					//Get a list of the corners
					float maxWallLength = DecentAgentUtil.getMaxWallSize(state.walls);
					var corners = DecentAgentUtil.findCornersTankIsInsideOf(state.walls, mypos, maxWallLength, false);

					//Get the closest corner
					corners.Sort((x, y) => ((x.first - mypos).magnitude).CompareTo((y.first - mypos).magnitude));

					//Check if we are on the inside of the corner
					if (corners != null && corners.Count > 0)
					{
						var corner = corners[0];

						var vec2corner = corner.first - mypos;
						var angleToCorner = Math.Atan2(vec2corner.y, vec2corner.x).RadiansToDegrees();
						if (angleToCorner < 0)
							angleToCorner += 360;

						//Get the closest outside endpoint of the corner walls
						var wallEnd = (corner.first.getNormalizedVectorPointingToEnd(corner.second.second) - mypos).magnitude < (corner.first.getNormalizedVectorPointingToEnd(corner.third.second) - mypos).magnitude ? corner.second.second : corner.third.second;

						var distToWall = Util.distToLineSegment(state.self.getPosition(), corner.first, wallEnd);
						//Debugging
						if (debuggingLines)
						{
							var v = corner.first;
							var o1 = corner.second.second;
							var o2 = corner.third.second;

							Debug.DrawLine(v, v + (o1 - v).normalized * maxWallLength, Color.black);
							Debug.DrawLine(v, v + (o2 - v).normalized * maxWallLength, Color.black);
							Debug.DrawLine(v + (o1 - v).normalized * maxWallLength, v + (o2 - v).normalized * maxWallLength, Color.black);
							Debug.Log(DecentAgentUtil.PointInTriangle(mypos, v, v + (o1 - v).normalized * maxWallLength, v + (o2 - v).normalized * maxWallLength));

							Debug.DrawRay(v, (mypos - v).normalized * longerBuffer, Color.blue);
							Debug.DrawRay(v, (mypos - v).normalized * shorterBuffer, Color.red);

						}

						//If in a corner
						if ((corner.first - mypos).magnitude <= longerBuffer)
						{
							//Find and set the priority of this action
							float priority = (higherPriority - lowerPriority) / (longerBuffer - shorterBuffer) * (mypos - corner.first).magnitude;
							action.SetPriority(priority, voteSourceID);

							//Get the angles to various points
							var vec2o = wallEnd.normalized * maxWallLength - mypos;
							var angleToWallEnd = Math.Atan2(vec2o.y, vec2o.x).RadiansToDegrees();
							if (angleToWallEnd < 0)
								angleToWallEnd += 360;

							var angleVehicle = Math.Atan2(state.self.getVelocity().y, state.self.getVelocity().x).RadiansToDegrees();
							if (angleVehicle < 0)
								angleVehicle += 360;

							float trackRight;

							//Corner is on the left side
							if (Mathf.Abs(angleToCorner) > 70 && Mathf.Abs(angleToCorner) < 290)
							{
								if (angleVehicle > angleToCorner)
								{   //Need to turn left
									trackRight = -1f;
								}
								else
								{   //Need to turn right
									trackRight = 1f;
								}
							}
							//Corner is on the right side. These must be dealt with differently because of 0-360 crossings
							else
							{
								//Bottom right corner
								if (angleToCorner > 180)
								{
									angleToCorner = -360 + angleToCorner;
									angleVehicle = angleVehicle > 180 ? -360 + angleVehicle : angleVehicle;

									if (angleVehicle < angleToCorner)
									{   //Need to turn right
										trackRight = 1f;
									}
									else
									{   //Need to turn left
										trackRight = -1f;
									}
								}
								//Top right corner
								else
								{
									angleVehicle = angleVehicle > 180 ? -360 + angleVehicle : angleVehicle;

									if (angleVehicle > angleToCorner)
									{   //Need to turn left
										trackRight = -1f;
									}
									else
									{   //Need to turn right
										trackRight = 1f;
									}
								}
							}

							//Set the action
							action.SetTurnTracksRight(turnRightMultiplier * trackRight, voteSourceID);
						}
						//Else get perpendicular to the wall that is part of the corner, preferably some distance away
						else if (distToWall.first <= (internalBuffer + wallBuffer))
						{
							//Get the trackRight value to use to get parallel to the wall
							float trackRight = DecentAgentUtil.findShortestTrackRotationPerpendicularToTarget(state.self.getPosition(), distToWall.second, state.self.getTracksRotation());

							//Used for debugging
							//Debug.Log("Track right setting: " + trackRight);
							if (debuggingLines)
							{
								Debug.DrawLine(state.self.getPosition(), distToWall.second, Color.green);
							}

							if (distToWall.first <= prefferedDistanceToWall)
							{
								//Determine if wall is on left or right
								Vector2 vecToWall = distToWall.second - mypos;
								float angleToWall = Math.Atan2(vecToWall.y, vecToWall.x).RadiansToDegrees();

								angleToWall -= state.self.getTracksRotation();

								while (angleToWall < 0)
								{
									angleToWall += 360f;
								}

								trackRight += angleToWall > 180 ? -1f : 1f;
								if (debuggingLines)
								{
									Debug.DrawLine(distToWall.second, distToWall.second + (mypos - distToWall.second).normalized * prefferedDistanceToWall, Color.red);
								}
							}

							//Set the action
							action.SetTurnTracksRight(trackRight, voteSourceID);

							//Set the variable priority
							if (distToWall.first < prefferedDistanceToWall)
								action.SetPriority(perpendicularToWallPriority, voteSourceID);
							else
								action.SetPriority(perpendicularToWallPriority - (prefferedDistanceToWall - distToWall.first), voteSourceID);
						}
						//Else get perpendicular to any wall if close
						else
						{
							var wall = DecentAgentUtil.getClosestWall(mypos, state.walls);
							distToWall = Util.distToLineSegment(mypos, wall.first, wall.second);
							if (distToWall.first < (normalWallBuffer + internalBuffer))
							{
								//Get the trackRight value to use to get parallel to the wall
								float trackRight = DecentAgentUtil.findShortestTrackRotationPerpendicularToTarget(state.self.getPosition(), distToWall.second, state.self.getTracksRotation());

								//Set the action
								action.SetTurnTracksRight(trackRight, voteSourceID);
								action.SetPriority(perpendicularToWallPriority, voteSourceID);
							}
						}
					}

					return action;
				}

				/// <summary>
				/// Write the values to the dictionary
				/// </summary>
				/// <param name="dict"></param>
				/// <param name="saveLastAction"></param>
				public override void _ToDict(Dictionary<string, object> dict, bool saveLastAction)
				{
					dict["HigherPriority"] = higherPriority;
					dict["ShorterBuffer"] = shorterBuffer;
					dict["LowerPriority"] = lowerPriority;
					dict["LongerBuffer"] = longerBuffer;
					dict["TurnRightMultiplier"] = turnRightMultiplier;
					dict["DebuggingLines"] = debuggingLines;
					dict["WallBuffer"] = wallBuffer;
					dict["NormalWallBuffer"] = normalWallBuffer;
					dict["PerpendicularToWallPriority"] = perpendicularToWallPriority;

				}

				/// <summary>
				/// Get the class values from the dictionary
				/// </summary>
				/// <param name="dict">Dictionary to restore from</param>
				public override void _FromDict(Dictionary<string, object> dict)
				{
					string bName = this.GetType().Name;
					higherPriority = Util.FloatFromJsonDict("HigherPriority", higherPriority, bName, dict);
					shorterBuffer = Util.FloatFromJsonDict("ShorterBuffer", shorterBuffer, bName, dict);
					lowerPriority = Util.FloatFromJsonDict("LowerPriority", lowerPriority, bName, dict);
					longerBuffer = Util.FloatFromJsonDict("LongerBuffer", longerBuffer, bName, dict);
					turnRightMultiplier = Util.FloatFromJsonDict("TurnRightMultiplier", turnRightMultiplier, bName, dict);
					wallBuffer = Util.FloatFromJsonDict("WallBuffer", wallBuffer, bName, dict);
					perpendicularToWallPriority = Util.FloatFromJsonDict("PerpendicularToWallPriority", perpendicularToWallPriority, bName, dict);
					normalWallBuffer = Util.FloatFromJsonDict("NormalWallBuffer", normalWallBuffer, bName, dict);
					debuggingLines = DecentAgentUtil.BoolFromJsonDict("DebuggingLines", debuggingLines, bName, dict);
				}
			}
		}
	}
}
