﻿using UnityEngine;
using System;
using System.Collections;
using System.Collections.Generic;

namespace AT
{
	namespace UBF
	{
		namespace Behaviors
		{
			/// <summary>
			/// Lets the tank go in a number of different patterns. Playing around with incrementTracksEachTime
			/// </summary>
			public class DecentAgent_Wander : AT.UBF.Behavior
			{
				/// <summary>
				/// Priority when there aren't any bullets of enemy tanks in sight
				/// </summary>
				public float priority = 1.0f;

				/// <summary>
				/// The amount to increment the turnTracksRight property at each time step.
				/// Keep in mind that the time step is small, so this value should be small as well (less than 0.1)
				/// A value of 0.01f is not a bad value, it makes the tank go in a really curvy line
				/// </summary>
				public float incrementTracksEachTime = 0.01f;

				/// <summary>
				/// Desired speed for the tank
				/// </summary>
				public float desiredSpeed = 0.8f;

				/// <summary>
				/// Internal state memory to hold the last turnTracksRight value. This is needed in order to not have a random steer.
				/// </summary>
				private float previousTracksTurnRight = 1.0f;

				/// <summary>
				/// Generate the actual action associated with the behavior
				/// </summary>
				/// <param name="state">State of the world as seen by the tank</param>
				/// <returns>Action appropriate for the given state according to this behavior</returns>
				protected override AT.UBF.Action genActionInternal(AT.UBF.State state)
				{
					AT.UBF.Action action = new AT.UBF.Action();

					List<string> originalMessages;
					var cstate = DecentAgentUtil.GetCombinedStateFromMessages(state, out originalMessages);

					if(cstate.combinedVisibleBullets.Count==0)
					{
						previousTracksTurnRight = previousTracksTurnRight + incrementTracksEachTime;
						//If the turnTrackRight value is above 1, then wrap it around to -1
						if (previousTracksTurnRight > 1)
							previousTracksTurnRight = -1;

						//Similarly, wrap if below -1
						if (previousTracksTurnRight < -1)
							previousTracksTurnRight = 1;

						//Set the action
						action.SetTurnTracksRight(previousTracksTurnRight, voteSourceID);
						action.SetPriority(priority, voteSourceID);
					}

					//Always be driving forward at least
					action.SetDriveForward(desiredSpeed, voteSourceID);

					return action;
				}

				/// <summary>
				/// Save to the dictionary
				/// </summary>
				/// <param name="dict"></param>
				/// <param name="saveLastAction"></param>
				public override void _ToDict(Dictionary<string, object> dict, bool saveLastAction)
				{
					dict["Priority"] = priority;
					dict["IncrementTracksEachTime"] = incrementTracksEachTime;
					dict["PreviousTracksTurnRight"] = previousTracksTurnRight;
					dict["DesiredSpeed"] = desiredSpeed;
				}

				/// <summary>
				/// Load from the dictionary
				/// </summary>
				/// <param name="dict">Dictionary</param>
				public override void _FromDict(Dictionary<string, object> dict)
				{
					string bName = this.GetType().Name;
					priority = Util.FloatFromJsonDict("Priority", priority, bName, dict);
					previousTracksTurnRight = Util.FloatFromJsonDict("PreviousTracksTurnRight", previousTracksTurnRight, bName, dict);
					incrementTracksEachTime = Util.FloatFromJsonDict("IncrementTracksEachTime", incrementTracksEachTime, bName, dict);
					desiredSpeed = Util.FloatFromJsonDict("DesiredSpeed", desiredSpeed, bName, dict);
				}
			}
		}
	}
}