﻿using UnityEngine;
using UnityEditor;
using System.Collections;
using System.Collections.Generic;
using System;
using System.IO;
using System.Text;

public static class DecentAgentUtil
{
	/// <summary>
	/// Helper extension method to get a vector perpendicular to the given vector
	/// </summary>
	/// <param name="myPos">Vector to find another vector that is perpendicular to it</param>
	/// <returns>Prependicular vector to the input</returns>
	public static Vector2 findPerpendicularVector(this Vector2 myPos)
	{
		return new Vector2(myPos.x, -myPos.y);
	}

	/// <summary>
	/// Returns the closest bullet that is coming in the general direction of mypos
	/// Note: This method does not see bullets that are close to the tank, but outside of the cone of visibility
	/// </summary>
	/// <param name="mypos">Position of the tank</param>
	/// <param name="bullets">List of visible bullets</param>
	/// <param name="buffer">Does not consider bullets inside this range. Defualts to 0.</param>
	/// <returns>Closest bullet incoming to the tank</returns>
	public static Tuple<Vector2, Vector2> getClosestIncomingBullet(this Vector2 mypos, List<Tuple<Vector2, Vector2>> bullets, float buffer = 2, bool debug = false)
	{
		List<Tuple<Vector2, Vector2>> incoming = getIncomingBullets(mypos, bullets, buffer, debug);

		if (incoming.Count > 0)
		{
			//Sorts the bullets into the closest to furthest away
			incoming.Sort((x, y) => ((x.first - mypos).magnitude).CompareTo((y.first - mypos).magnitude));

			//Return the closest bullet
			return incoming[0];
		}
		return null;
	}

	/// <summary>
	/// Projects a position vector onto a line specified by a position and velocity
	/// </summary>
	/// <param name="tank">Tuple containing the position and velocity of the tank, in that order</param>
	/// <param name="bullet">Position vector to the bullet</param>
	/// <returns>Position vector to the point on the line specified by tank closest to the bullet</returns>
	public static Vector2 projectVectortoLine(Tuple<Vector2, Vector2> tank, Vector2 bullet)
	{
		if (Math.Abs(tank.second.magnitude) <= 1e-13)
			return new Vector2(tank.first.x, tank.first.y);
		if (Math.Abs(tank.second.x) <= Math.Pow(10, -13))
			return new Vector2(tank.first.x, bullet.y);
		if (Math.Abs(tank.second.y) <= Math.Pow(10, -13))
		{
			return new Vector2(bullet.x, tank.first.y);
		}


		float m = tank.second.y / tank.second.x;
		float b = tank.first.y - m * tank.first.x;
		float x = (m * bullet.y + bullet.x - m * b) / (m * m + 1);
		float y = (m * m * bullet.y + m * bullet.x + b) / (m * m + 1);
		return new Vector2(x, y);
	}

	/// <summary>
	/// Gets the intersection point of two lines
	/// </summary>
	/// <param name="tank">Tuple containing the position and velocity of the tank, in that order</param>
	/// <param name="bullet">Tuple containing the position and velocity of the bullet, in that order</param>
	/// <returns>Position vector to the point on the line specified by the tank closest to the bullet</returns>
	public static Vector2 getIntersectionPoint(Tuple<Vector2, Vector2> t, Tuple<Vector2, Vector2> b)
	{
		if (Math.Abs(t.second.magnitude) <= 2e-13)
			return new Vector2(t.first.x, t.first.y);
		if (Math.Abs(t.second.x) <= 1e-13)
			return new Vector2(t.first.x, b.first.y);
		if (Math.Abs(t.second.y) <= 1e-13)
			return new Vector2(b.first.x, t.first.y);

		float mt = t.second.y / t.second.x;
		float bt = t.first.y - mt * t.first.x;
		float mb = b.second.y / b.second.x;
		float bb = b.first.y - mb * b.first.x;
		float x = (bt - bb) / (mb - mt);
		float y = mt * x + bt;
		return new Vector2(x, y);
	}

	/// <summary>
	/// Gets the closest wall to mypos
	/// </summary>
	/// <param name="mypos">Position of the tank</param>
	/// <param name="walls">List of walls in the game</param>
	/// <returns>Wall that is closest to the tank</returns>
	public static Tuple<Vector2, Vector2> getClosestWall(Vector2 mypos, List<Tuple<Vector2, Vector2>> walls)
	{
		Tuple<Vector2, Vector2> closestWall = walls[0];
		Tuple<float, Vector2> temp, closest = Util.distToLineSegment(mypos, walls[0].first, walls[0].second);

		foreach (var wall in walls)
		{
			temp = Util.distToLineSegment(mypos, wall.first, wall.second);
			if (temp.first < closest.first)
			{
				closest = temp;
				closestWall = wall;
			}
		}
		return closestWall;
	}

	/// <summary>
	/// Finds turnRight value to turn the tracks until the tank is perpendicular to the target posistion
	/// </summary>
	/// <param name="myPos">Position of the tank of interest</param>
	/// <param name="targetPos">Position of the target to be perdpendicular to</param>
	/// <param name="myTrackRotation">Angle of the tank of interest</param>
	/// <returns>Rotation speed to turn the tank to be perpendicular to the target position</returns>
	public static float findShortestTrackRotationPerpendicularToTarget(Vector2 myPos, Vector2 targetPos, float myTrackRotation, bool debuggingLines = false)
	{
		float targetRotation1 = Util.getRelativeAngleToTarget(myPos, targetPos) + 90;
		targetRotation1 = targetRotation1 >= 360 ? targetRotation1 - 360f : targetRotation1;
		targetRotation1 = targetRotation1 > 180 ? -(360 - targetRotation1) : targetRotation1;
		//targetRotation1 = targetRotation1 < 0 ? targetRotation1 + 360f : targetRotation1;

		float targetRotation2 = Util.getRelativeAngleToTarget(myPos, targetPos) - 90;
		targetRotation2 = targetRotation2 > 180 ? -(360f - targetRotation2) : targetRotation2;
		//targetRotation2 = targetRotation2 < 0 ? targetRotation2 + 360f : targetRotation2;
		//targetRotation2 = targetRotation2 >= 180 ? targetRotation2 - 180f : targetRotation2;

		float targetRotation;


		var temp1 = myTrackRotation - targetRotation1;
		var temp2 = myTrackRotation - targetRotation2;

		temp1 = temp1 > 180 ? -(360f - temp1) : temp1;
		temp2 = temp2 > 180 ? -(360f - temp2) : temp2;

		if (Math.Abs(temp1) % 360 < Math.Abs(temp2) % 360)
			targetRotation = targetRotation1;
		else
			targetRotation = targetRotation2;


		if (targetRotation > 180)
			targetRotation = -(360 - myTrackRotation);
		if (debuggingLines)
			Debug.DrawLine(myPos, myPos + (new Vector2(Mathf.Cos(targetRotation), Mathf.Sin(targetRotation))) * 2, Color.blue);

		//DecentAgentUtil.DrawRayInDirection(myPos, targetRotation, Color.green);
		float maxRot = AT.Tank.trackRotationSpeed;
		return findShortestRotation(targetRotation, myTrackRotation, maxRot);
	}

	/// <summary>
	/// Draws a ray on the screen from the given location into a certain direction
	/// </summary>
	/// <param name="pos">Starting position of the ray</param>
	/// <param name="dir">Direction to point the ray in</param>
	/// <param name="color">Color of the ray</param>
	public static void DrawRayInDirection(Vector2 pos, float dir, Color color)
	{
		Vector2 dirvec = new Vector2((float)Math.Cos(dir.DegreesToRadians()), (float)Math.Sin(dir.DegreesToRadians()));
		Debug.DrawRay(pos, dirvec, color);
	}

	/// <summary>
	/// Extension method to convert degrees to radians
	/// </summary>
	/// <param name="Angle in degrees">Degrees</param>
	/// <returns>Same angle in radians</returns>
	public static double DegreesToRadians(this float deg)
	{
		return (double)deg * Math.PI / 180;
	}

	/// <summary>
	/// Extension method to convert radians to degrees
	/// </summary>
	/// <param name="rad">Radians</param>
	/// <returns>Same angle in degrees</returns>
	public static float RadiansToDegrees(this double rad)
	{
		return (float)(rad / Math.PI * 180);
	}

	/// <summary>
	/// Generates a message a bit easier than built in.
	/// </summary>
	/// <param name="state">Overall state as visible from the tank</param>
	/// <param name="voteSourceID">Integer where the vote came from</param>
	/// <param name="message">Message to be sent to the team</param>
	/// <returns>Message instance that just needs to be sent</returns>
	public static AT.Message GenerateMessage(AT.UBF.State state, int voteSourceID, string message)
	{
		string sender = state.self.getColor();
		string myTeam = state.self.getTeam().ToString(); // enum to string representation
		message += Environment.NewLine + toCSV(state);
		AT.Message msg = new AT.Message(voteSourceID, sender, AT.Message.ReceiverType.Team, myTeam, message);
		AT.Tank.SendMessage(msg);
		return msg;
	}

	/// <summary>
	/// Sends a message without adding any state representation to the end. This can be useful for commands
	/// </summary>
	/// <param name="state">State of the environment as seen by the agent. Note that this is not included in the message</param>
	/// <param name="voteSourceID">Integer where the vote came from</param>
	/// <param name="message">Message to be sent to the team</param>
	public static void SendNonStateMesssage(AT.UBF.State state, int voteSourceID, string message)
	{
		string sender = state.self.getColor();
		string myTeam = state.self.getTeam().ToString(); // enum to string representation
		AT.Message msg = new AT.Message(voteSourceID, sender, AT.Message.ReceiverType.Team, myTeam, message);
		AT.Tank.SendMessage(msg);
	}

	/// <summary>
	/// This method was pulled from Util.cs because it is a private method there. The goal of this class is to provide the helper methods needed in my created methods.
	/// </summary>
	/// <param name="targetRotation">Target angle in degrees</param>
	/// <param name="currRotation">Current angle in degrees</param>
	/// <param name="maxRotation">Maximum rotation in one time step</param>
	/// <returns>turnRight value that will not overshoot the target rotation before the next time step</returns>
	private static float findShortestRotation(float targetRotation, float currRotation, float maxRotation)
	{
		float temp = (targetRotation - currRotation + 360) % 360;
		float turnRight = 0.0f;

		if (temp > 180)
		{
			if (temp > 360 - maxRotation)
			{
				turnRight = Mathf.Abs(360 - temp) / maxRotation;
			}
			else
			{
				turnRight = 1.0f;
			}
		}
		else
		{
			if (temp < maxRotation)
			{
				turnRight = (0 - temp) / maxRotation;
			}
			else
			{
				turnRight = -1.0f;
			}
		}

		return turnRight;
	}

	/// <summary>
	/// Gets the overall state as seen by all of the tanks on a team
	/// </summary>
	/// <param name="state">State seen by the current tank</param>
	/// <param name="originalMessages">List of original messages that do not include state information, only commands</param>
	/// <returns>The combined state of the environment as seen through the eyes of every tank on the team</returns>
	public static CombinedState GetCombinedStateFromMessages(AT.UBF.State state, out List<string> originalMessages)
	{
		//Initialize the combined state
		CombinedState c = new CombinedState();
		c.combinedVisibleBullets = new List<Tuple<Vector2, Vector2>>(state.visibleBullets);

		//Convert and add the visible enemy tanks to the visible tank list
		c.combinedVisibleEnemyTanks = new List<CombinedState.OtherTankState>();
		foreach (var tank in state.visibleEnemyTanks)
		{
			if (tank.getTeam() != state.self.getTeam())
			{
				CombinedState.OtherTankState enemy = new CombinedState.OtherTankState();
				enemy.color = tank.getColor();
				enemy.team = tank.getTeam();
				enemy.position = tank.getPosition();
				enemy.velocity = tank.getVelocity();
				enemy.tracksRotation = tank.getTracksRotation();
				enemy.turretRotation = tank.getTurretRotation();
				enemy.health = tank.getHealth();
				c.combinedVisibleEnemyTanks.Add(enemy);
			}
		}
		c.walls = new List<Tuple<Vector2, Vector2>>(state.walls);
		c.teamates = new List<CombinedState.OtherTankState>();
		originalMessages = new List<string>();
		foreach (var message in state.messages)
		{
			originalMessages.Add(fromCSV(state, message.getMessage(), ref c));
		}

		c.teamates.RemoveAll(x => x.color == state.self.getColor());

		return c;
	}

	/// <summary>
	/// Converts a message in csv form to its equivalent combined state and extracts the original message
	/// </summary>
	/// <param name="currentState">Current state of the world as seen by the calling tank</param>
	/// <param name="csv">Message straight from the source</param>
	/// <param name="c">CombinedState of the world so far. Does not include the state as seen by the sender of this message at this timestep</param>
	/// <param name="originalMessage">Original (non stat-related message)</param>
	/// <returns>Combined state updated with what the sender of the message can see</returns>
	private static string fromCSV(AT.UBF.State currentState, string csv, ref CombinedState c)
	{
		//Initialize a stringbuilder
		StringBuilder message = new StringBuilder();

		bool foundStateRepresentation = false;

		//Get the other tank's state
		CombinedState.OtherTankState otherTank = new CombinedState.OtherTankState();
		string[] lines = csv.Split(new string[] { Environment.NewLine }, StringSplitOptions.RemoveEmptyEntries);
		int i = 0;
		for (i = 0; i < lines.Length; i++)
		{
			if (lines[i].CompareTo("State:") == 0)
			{
				foundStateRepresentation = true;
				break;
			}
			else
			{
				if (lines[i].Length != 0)
				{
					message.Append(lines[i] + Environment.NewLine);
				}
			}
		}

		//If there is no state representation, then just get the
		if (!foundStateRepresentation)
		{
			return message.ToString();
		}

		//TODO: Might need an i++ here
		i++;

		//Get the other tank state
		otherTank = new CombinedState.OtherTankState();
		otherTank.team = lines[i++].Substring(6).ToEnum();
		otherTank.color = lines[i++].Substring(7);
		otherTank.position = getVector2FromLine(lines[i++]);
		otherTank.velocity = getVector2FromLine(lines[i++]);
		otherTank.turretRotation = getFloatFromLine(lines[i++]);
		otherTank.health = getIntFromLine(lines[i++]);
		otherTank.tracksRotation = getFloatFromLine(lines[i++]);

		//Add it to the otherTank state list
		c.teamates.Add(otherTank);


		//Extract the enemy tanks
		//First, find the first line of the enemy tanks
#pragma warning disable CS1717 // Assignment made to same variable
		for (i = i; i < lines.Length; i++)
#pragma warning restore CS1717 // Assignment made to same variable
		{
			if (lines[i].CompareTo("Start of: Visible enemy tanks") == 0)
				break;
		}

		//TODO: Might need an i++ here
		i++;
		bool exists;
		string[] line;
		while (lines[i].CompareTo("End of: Visible enemy tanks") != 0)
		{
			//Get the other tank state
			otherTank = new CombinedState.OtherTankState();
			line = lines[i++].Split(new string[] { ";" }, StringSplitOptions.RemoveEmptyEntries);
			otherTank.team = line[0].Substring(14).ToEnum();
			otherTank.color = line[1].Substring(15);
			otherTank.position = getVector2FromLine(line[2]);
			otherTank.velocity = getVector2FromLine(line[3]);
			otherTank.turretRotation = getFloatFromLine(line[4]);
			otherTank.health = getIntFromLine(line[5]);
			otherTank.tracksRotation = getFloatFromLine(line[6]);

			//Check to make sure that this tank is not already present in the list
			exists = false;
			foreach (var tank in c.combinedVisibleEnemyTanks)
			{
				if (tank.IsEqual(otherTank))
				{
					exists = true;
					break;
				}
			}

			if (!exists && otherTank.team != currentState.self.getTeam())
			{
				//Add it to the otherTank state list
				c.combinedVisibleEnemyTanks.Add(otherTank);
			}
		}


		//Now do the same for the bullets

		//First, find the first line of the enemy tanks
#pragma warning disable CS1717 // Assignment made to same variable
		for (i = i; i < lines.Length; i++)
#pragma warning restore CS1717 // Assignment made to same variable
		{
			if (lines[i].CompareTo("Start of: Visible bullets") == 0)
				break;
		}

		//TODO: might need an i++ here
		i++;
		Vector2 pos, vel;
		Tuple<Vector2, Vector2> otherBullet;
		while (lines[i].CompareTo("End of: Visible bullets") != 0)
		{
			line = lines[i++].Split(new string[] { ";" }, StringSplitOptions.RemoveEmptyEntries);
			pos = getVector2FromLine(line[0]);
			vel = getVector2FromLine(line[1]);

			otherBullet = new Tuple<Vector2, Vector2>(pos, vel);
			//Check to make sure that this tank is not already present in the list
			exists = false;
			foreach (var bullet in c.combinedVisibleBullets)
			{
				if (bullet.BulletIsEqual(otherBullet))
				{
					exists = true;
					break;
				}
			}

			if (!exists)
			{
				//Add it to the otherTank state list
				c.combinedVisibleBullets.Add(otherBullet);
			}
		}



		return message.ToString();
	}

	/// <summary>
	/// Converts all possible knowledge about the state of the environment to a string that can be passed to the team via a message
	/// </summary>
	/// <param name="state">State as seen by one tank</param>
	/// <returns>String containing all useful information about the state</returns>
	private static string toCSV(AT.UBF.State state)
	{
		StringBuilder str = new StringBuilder("State:" + Environment.NewLine);

		//Add the team of the sender
		str.Append("Team: " + state.self.getTeam() + Environment.NewLine);

		//Add the color of the sender
		str.Append("Color: " + state.self.getColor() + Environment.NewLine);

		//Add the tanks current location
		str.Append("Current Position: " + state.self.getPosition() + Environment.NewLine);

		//Add the tanks current velocity
		str.Append("Current Velocity: " + state.self.getVelocity() + Environment.NewLine);

		//Add the tanks current turret rotation
		str.Append("Current Turret Rotation: " + state.self.getTurretRotation() + Environment.NewLine);

		//Add the tanks current health
		str.Append("Current Health: " + state.self.getHealth() + Environment.NewLine);

		//Add the tanks current tracks rotation
		str.Append("Current Tracks Rotation: " + state.self.getTracksRotation() + Environment.NewLine);

		//Add the list of visible enemy tanks
		str.Append("Start of: Visible enemy tanks" + Environment.NewLine);
		foreach (var tank in state.visibleEnemyTanks)
		{
			str.Append("Current Team: " + tank.getTeam() +
				";Current Color: " + tank.getColor() +
				";Current Position: " + tank.getPosition() +
				";Current Velocity: " + tank.getVelocity() +
				";Current Turret Rotation: " + state.self.getTurretRotation() +
				";Current Health: " + state.self.getHealth() +
				";Current Tracks Rotation: " + state.self.getTracksRotation() + Environment.NewLine);
		}
		str.Append("End of: Visible enemy tanks" + Environment.NewLine);


		str.Append("Start of: Visible bullets" + Environment.NewLine);
		foreach (var bullet in state.visibleBullets)
		{
			str.Append("Bullet position: " + bullet.first +
				";Bullet velocity: " + bullet.second + Environment.NewLine);
		}
		str.Append("End of: Visible bullets" + Environment.NewLine);
		return str.ToString();
	}

	/// <summary>
	/// Gets a vector from a line containing the format (x,y) exactly once
	/// </summary>
	/// <param name="line">String of the format described above to extract a vector from</param>
	/// <returns>Vector2 instance describing the string</returns>
	private static Vector2 getVector2FromLine(string line)
	{
		float x = 0, y = 0;

		int index = line.IndexOf('(');
		string vector = line.Substring(index + 1).Replace(")", "");
		string[] xy = vector.Split(new char[] { ',' });
		if (!float.TryParse(xy[0], out x))
		{
			x = float.NaN;
		}

		if (!float.TryParse(xy[1], out y))
			y = float.NaN;

		return new Vector2(x, y);
	}

	/// <summary>
	/// Gets a floating point argument from the line given
	/// </summary>
	/// <param name="line">Line string containing an integer in the following format:
	/// ~string don't care~: floathere,</param>
	/// <returns>Integer from the line in int form</returns>
	private static float getFloatFromLine(string line)
	{
		float x = 0;

		int index = line.IndexOf(':');
		string f = line.Substring(index + 1).Replace(",", "").Trim();
		if (!float.TryParse(f, out x))
		{
			x = float.NaN;
		}
		return x;
	}

	/// <summary>
	/// Gets an integer argument from the line given
	/// </summary>
	/// <param name="line">Line string containing an integer in the following format:
	/// ~string don't care~: integerhere,</param>
	/// <returns>float from the line in float form</returns>
	private static int getIntFromLine(string line)
	{
		int x = 0;

		int index = line.IndexOf(':');
		string f = line.Substring(index + 1).Replace(",", "").Trim();
		if (!int.TryParse(f, out x))
		{
			x = int.MinValue;
		}
		return x;
	}

	/// <summary>
	/// an instance of this class can hold information about the state of the map combined from multiple different tanks. This allows the tank to make decisions based on more information
	/// </summary>
	public class CombinedState
	{

		public class OtherTankState
		{
			public AT.Team team;
			public string color;
			public Vector2 position;
			public Vector2 velocity;
			public float turretRotation;
			public float tracksRotation;
			public int health;
		}

		//List of the other tanks on your team
		public List<OtherTankState> teamates;

		public List<OtherTankState> combinedVisibleEnemyTanks;
		// first vector is position, second is velocity
		public List<Tuple<Vector2, Vector2>> combinedVisibleBullets;
		//All wall positions are known all the time
		public List<Tuple<Vector2, Vector2>> walls;
	}

	/// <summary>
	/// Compares the current Vector2 Vector2 pair to another
	/// </summary>
	/// <param name="cur">Current tuple</param>
	/// <param name="other">Other tuple</param>
	/// <returns>True if they are the same, false if not</returns>
	public static bool BulletIsEqual(this Tuple<Vector2, Vector2> cur, Tuple<Vector2, Vector2> other)
	{
		if (cur.first.PastLocationAboutEqual(cur.second, other.first, 0.1f) &&
			cur.second.x == other.second.x &&
			cur.second.y == other.second.y)
		{
			return true;
		}

		return false;
	}

	private static bool PastLocationAboutEqual(this Vector2 currentPos, Vector2 velocity, Vector2 past, float eps=0.1f)
	{
		var roughPast = currentPos - velocity * Time.deltaTime;
		if (Mathf.Abs(roughPast.x - past.x) < eps &&
			Mathf.Abs(roughPast.y - past.y) < eps)
			return true;
		return false;

	}

	/// <summary>
	/// Compares the current tank to another tank
	/// </summary>
	/// <param name="cur">Current tank</param>
	/// <param name="other">Other tank to compare</param>
	/// <returns>True if they are the same, false if not</returns>
	public static bool IsEqual(this CombinedState.OtherTankState cur, CombinedState.OtherTankState other)
	{
		//Position is not included in this because if the tank is moving it will not be exactly the same
		//Unfortanately, this assumes that there is only one tank of each color per team
		if (cur.team == other.team &&
			cur.color == other.color)
		{
			return true;
		}

		return false;
	}

	/// <summary>
	/// Converts the string representation of an enum to an actual enum value
	/// </summary>
	/// <param name="stringValue">String to convert to a team</param>
	/// <returns>Enum representation of the string</returns>
	public static AT.Team ToEnum(this string stringValue)
	{
		if (!Enum.IsDefined(typeof(AT.Team), stringValue))
			return AT.Team.None;

		return (AT.Team)Enum.Parse(typeof(AT.Team), stringValue);
	}

	/// <summary>
	/// Gets the bullets that are coming at the tank to within a distance of buffer
	/// </summary>
	/// <param name="bullets">List of visible bullets</param>
	/// <param name="buffer">Distance away from the tank to not care about bullets that fall outside of this range.
	/// This distance is along a line perpendicular to the bullets velocity through the ceter of the tank</param>
	/// <returns>List of bullets that are currently set to impact the line perpendicular to the bullets velocity within buffer of the tank.</returns>
	public static List<Tuple<Vector2, Vector2>> getIncomingBullets(Vector2 mypos, List<Tuple<Vector2, Vector2>> bullets, float buffer, bool debug = false)
	{
		List<Tuple<Vector2, Vector2>> incoming = new List<Tuple<Vector2, Vector2>>();
		Vector2 impactPointPA;
		float relAngle;

		foreach (var bullet in bullets)
		{
			//Ensure that the bullet is actually coming in
			if ((bullet.first + bullet.second * 0.00001f - mypos).magnitude >= (bullet.first - mypos).magnitude)
			{
				continue;
			}

			//Get the realtive angle from the bullets velocity and position to the tanks centerline
			relAngle = getAngle(mypos, bullet.first, bullet.first + bullet.second);

			//Use that angle to find the point at which the bullet will be closest to the tank
			impactPointPA = bullet.first + bullet.second.normalized * ((float)Math.Cos(relAngle.DegreesToRadians()) * (mypos - bullet.first).magnitude);

			//Ensure that the bullet is going to hit the line perpendicular to the bullet and through the center of the tank within the buffer
			if ((impactPointPA - mypos).magnitude > buffer)
			{
				continue;
			}

			if (debug)
			{
				Debug.DrawRay(mypos, buffer * (impactPointPA - mypos).normalized, Color.cyan);
				Debug.DrawLine(mypos, bullet.first, Color.red);
				Debug.DrawLine(bullet.first, impactPointPA, Color.white);
				Debug.DrawLine(mypos, impactPointPA, Color.white);
			}

			incoming.Add(bullet);
		}

		return incoming;
	}

	/// <summary>
	/// Gets the angle between three vectors with vector 2 being the vertex
	/// </summary>
	/// <param name="one">Vector one</param>
	/// <param name="two">Vertex of the angle</param>
	/// <param name="three">Vector three</param>
	/// <returns></returns>
	public static float getAngle(Vector2 one, Vector2 two, Vector2 three)
	{
		float side12 = (one - two).magnitude;
		float side13 = (one - three).magnitude;
		float side23 = (two - three).magnitude;
		float angle = lawOfCosines(side23, side13, side12);
		while (angle < 0)
		{
			angle += 360;
		}
		while (angle > 360)
		{
			angle -= 360;
		}
		return angle;
	}

	/// <summary>
	/// Uses the law of cosines to determine the size of the angle across from side B
	/// </summary>
	/// <param name="A">Length of side A</param>
	/// <param name="B">Length of side B</param>
	/// <param name="C">Length of side C</param>
	/// <returns>The angle of the angle opposite from side B</returns>
	private static float lawOfCosines(float a, float b, float c)
	{
		return Math.Acos((a * a + c * c - b * b) / (2 * a * c)).RadiansToDegrees();
	}

	/// <summary>
	/// Returns the closest enemy tank not behind a wall, or the closest enemy behind a wall if there are none in sight
	/// </summary>
	/// <param name="self">The self tank</param>
	/// <param name="tanks">List of tanks</param>
	/// <returns>Closest tank in not behind a wall, or if there are none of those, then the closest behind a wall</returns>
	public static CombinedState.OtherTankState getClosestEnemyTank(AT.UBF.State state, List<CombinedState.OtherTankState> tanks, bool debug = false)
	{
		if (tanks.Count > 0)
		{
			CombinedState.OtherTankState closestBH = tanks[0], closestIS = null;

			foreach (var tank in tanks)
			{
				if (closestIS == null && !checkIfWallInBetween(state.self.getPosition(), tank.position, state.walls))
				{
					closestIS = tank;
				}
				if (closestIS != null && !checkIfWallInBetween(state.self.getPosition(), tank.position, state.walls))
				{
					if ((state.self.getPosition() - tank.position).magnitude < (closestIS.position - state.self.getPosition()).magnitude
					&& tank.team != state.self.getTeam())
					{
						closestIS = tank;
					}
				}
				else if ((state.self.getPosition() - tank.position).magnitude < (closestBH.position - state.self.getPosition()).magnitude
					&& tank.team != state.self.getTeam())
				{
					closestBH = tank;
				}
			}

			closestIS = closestIS == null ? closestBH : closestIS;

			if (debug)
				Debug.DrawLine(state.self.getPosition(), closestIS.position, Color.cyan);

			return closestIS;
		}
		return null;
	}

	/// <summary>
	/// Checks if there is a wall between the given tank position and the target position
	/// </summary>
	/// <param name="mypos">Position of the tank</param>
	/// <param name="target">Other position to check if there is a wall in between</param>
	/// <param name="walls">List of all the walls in the map</param>
	/// <returns>True if there is a wall in between, false if not</returns>
	public static bool checkIfWallInBetween(Vector2 mypos, Vector2 target, List<Tuple<Vector2, Vector2>> walls)
	{
		foreach (var wall in walls)
		{
			if (LineIntersection.doLinesIntersect(new Tuple<Vector2, Vector2>(mypos, target), wall))
			{
				return true;
			}
		}
		return false;
	}

	/// <summary>
	/// Finds all of the corners due to two walls ending at the same location
	/// </summary>
	/// <param name="walls">List of all of the walls in the map</param>
	/// <returns>List of corners in the map</returns>
	public static List<Triple<Vector2, Tuple<Vector2, Vector2>, Tuple<Vector2, Vector2>>> findCorners(List<Tuple<Vector2, Vector2>> walls)
	{
		var corners = new List<Triple<Vector2, Tuple<Vector2, Vector2>, Tuple<Vector2, Vector2>>>();
		for (int i = 0; i < walls.Count; i++)
		{
			for (int k = i + 1; k < walls.Count; k++)
			{
				//Add the places where two walls meet at a point
				if (walls[i].first == walls[k].first)
				{
					corners.Add(new Triple<Vector2, Tuple<Vector2, Vector2>, Tuple<Vector2, Vector2>>(walls[i].first, walls[i], walls[k]));
				}
				else if (walls[i].first == walls[k].second)
				{
					corners.Add(new Triple<Vector2, Tuple<Vector2, Vector2>, Tuple<Vector2, Vector2>>(walls[i].first, walls[i], new Tuple<Vector2, Vector2>(walls[k].second, walls[k].first)));
				}
				else if (walls[i].second == walls[k].first)
				{
					corners.Add(new Triple<Vector2, Tuple<Vector2, Vector2>, Tuple<Vector2, Vector2>>(walls[i].second, new Tuple<Vector2, Vector2>(walls[i].second, walls[i].first), walls[k]));
				}
				else if (walls[i].second == walls[k].second)
				{
					corners.Add(new Triple<Vector2, Tuple<Vector2, Vector2>, Tuple<Vector2, Vector2>>(walls[i].second, new Tuple<Vector2, Vector2>(walls[i].second, walls[i].first), new Tuple<Vector2, Vector2>(walls[k].second, walls[k].first)));
				}
			}
		}

		return corners;
	}

	/// <summary>
	/// Determines if the tank is in the corner or outside
	/// </summary>
	/// <param name="corner">Corner to check if the tank is in</param>
	/// <param name="mypos">Current position of the tank</param>
	/// <param name="maxWallLength">Maximum length of any wall on the map</param>
	/// <param name="debug">Prints lines on the scene window to help figure out what is going on</param>
	/// <returns>True if the tank is on the inside of the corner, and false if not</returns>
	public static bool CheckOnInsideOfCorner(Triple<Vector2, Tuple<Vector2, Vector2>, Tuple<Vector2, Vector2>> corner, Vector2 mypos, float maxWallLength = 100, bool debug = false)
	{
		Vector2 o1 = corner.second.second, o2 = corner.third.second, v = corner.first;
		//First figure out which Vectors correspond to the vertex of the corner


		if (debug)
		{
			Debug.DrawLine(v, v + (o1 - v).normalized * maxWallLength, Color.black);
			Debug.DrawLine(v, v + (o2 - v).normalized * maxWallLength, Color.black);
			Debug.DrawLine(v + (o1 - v).normalized * maxWallLength, v + (o2 - v).normalized * maxWallLength, Color.black);
			Debug.Log(PointInTriangle(mypos, v, v + (o1 - v).normalized * maxWallLength, v + (o2 - v).normalized * maxWallLength));
		}

		//return PointInTriangle(mypos, v, v + (o1 - v).normalized * maxWallLength, v + (o2 - v).normalized * maxWallLength);
		return PointInTriangle(mypos, v, o1, o2);
	}

	/// <summary>
	/// Function taken from http://stackoverflow.com/questions/2049582/how-to-determine-a-point-in-a-2d-triangle
	/// Accessed 7/26/2016
	/// Poster: Kornel Kisielewicz
	/// </summary>
	/// <param name="p1"></param>
	/// <param name="p2"></param>
	/// <param name="p3"></param>
	/// <returns></returns>
	private static float sign(Vector2 p1, Vector2 p2, Vector2 p3)
	{
		return (p1.x - p3.x) * (p2.y - p3.y) - (p2.x - p3.x) * (p1.y - p3.y);
	}

	/// <summary>
	/// Determines if a point is inside the given triangle
	/// Function taken from http://stackoverflow.com/questions/2049582/how-to-determine-a-point-in-a-2d-triangle
	/// Accessed 7/26/2016
	/// Poster: Kornel Kisielewicz
	/// </summary>
	/// <param name="pt">Point to test</param>
	/// <param name="v1">First vertex</param>
	/// <param name="v2">Second vertex</param>
	/// <param name="v3">Third vertex</param>
	/// <returns>True if point is inside triangle, false if not</returns>
	public static bool PointInTriangle(Vector2 pt, Vector2 v1, Vector2 v2, Vector2 v3)
	{
		bool b1, b2, b3;

		b1 = sign(pt, v1, v2) < 0.0f;
		b2 = sign(pt, v2, v3) < 0.0f;
		b3 = sign(pt, v3, v1) < 0.0f;

		return ((b1 == b2) && (b2 == b3));
	}

	/// <summary>
	/// Gets the longest wall on the map
	/// </summary>
	/// <param name="walls">List of walls in the map</param>
	/// <returns>Length of the longest wall on the map</returns>
	public static float getMaxWallSize(List<Tuple<Vector2, Vector2>> walls)
	{
		float longest = 0;

		foreach (var wall in walls)
		{
			if (longest < (wall.first - wall.second).magnitude)
			{
				longest = (wall.first - wall.second).magnitude;
			}
		}
		return longest;
	}

	/// <summary>
	/// Gets a list of the corners that the tank is on the inside of
	/// </summary>
	/// <param name="walls">List of walls in the map</param>
	/// <param name="mypos">Current tank position</param>
	/// <param name="maxWallLength">Maximum length of any wall in the map</param>
	/// <param name="debug">Prints pretty lines if you want to see them</param>
	/// <returns>List of the corners that the tank is inside of</returns>
	public static List<Triple<Vector2, Tuple<Vector2, Vector2>, Tuple<Vector2, Vector2>>> findCornersTankIsInsideOf(List<Tuple<Vector2, Vector2>> walls, Vector2 mypos, float maxWallLength = 100, bool debug = false)
	{
		var corners = findCorners(walls);
		for (int i = corners.Count - 1; i >= 0; i--)
		{
			if (!CheckOnInsideOfCorner(corners[i], mypos, maxWallLength, debug))
			{
				corners.RemoveAt(i);
			}
		}
		return corners;
	}

	/// <summary>
	/// Gets a new vector that is one unit away from the origin vector in the direction of the end vector
	/// </summary>
	/// <param name="origin">Point to offset from</param>
	/// <param name="end">Endpoint of the original vector</param>
	/// <returns>New point that is one unit away from origin in the direction of the end point</returns>
	public static Vector2 getNormalizedVectorPointingToEnd(this Vector2 origin, Vector2 end)
	{
		Vector2 direction = (end - origin).normalized;
		return new Vector2(origin.x + direction.x, origin.y + direction.y);
	}

	/// <summary>
	/// Get a bool value from the JSON dictionary
	/// </summary>
	/// <param name="keyName"></param>
	/// <param name="defaultValue"></param>
	/// <param name="behaviorName"></param>
	/// <param name="dict"></param>
	/// <returns></returns>
	public static bool BoolFromJsonDict(string keyName, bool defaultValue, string behaviorName, Dictionary<string, object> dict)
	{
		bool result = defaultValue;
		try
		{
			result = (bool)dict[keyName];
		}
		catch (KeyNotFoundException)
		{
			Debug.Log(String.Format("'{0}' key not found in {1} Arguments. Using default of {2}.", keyName, behaviorName, defaultValue));
		}
		catch (InvalidCastException)
		{
			Debug.LogError(String.Format("'{0}' value in {1} Arguments is not a valid float", keyName, behaviorName));
		}
		return result;
	}

	/// <summary>
	/// Checks if a teamate is in the firing line. It does not work very well, but the teamate is able to dodge, so it is mostly okay
	/// </summary>
	/// <param name="mypos">Position of thr firing tank</param>
	/// <param name="turretAngle">Angle of the turret</param>
	/// <param name="teamates">List of teamates</param>
	/// <returns>True if there is a teamate in the firing line, false if not</returns>
	public static bool checkTeamateInShootingLine(AT.UBF.State state, List<CombinedState.OtherTankState> teamates, bool debug=false)
	{
		//If there are no teamates, return false
		if (teamates == null || teamates.Count == 0)
		{
			return false;
		}
		var mycolor = state.self.getColor();
		var mypos = state.self.getPosition();
		var turretAngle = state.self.getTurretRotation();

		//Get the firing line
		var line = new Tuple<Vector2, Vector2>(mypos, mypos + (new Vector2((float)Math.Cos(turretAngle.DegreesToRadians()), (float)Math.Sin(turretAngle.DegreesToRadians()))) * 100);
		if (debug)
		{
			Debug.DrawLine(line.first, line.second, Color.cyan);
		}
		foreach (var teamate in teamates)
		{
			//Get the minimum distance from the firing line to the teamate
			var minDist = Util.distToLineSegment(teamate.position, line.first, line.second);
			if (debug)
			{
				Debug.DrawLine(teamate.position, minDist.second, Color.blue);
			}

			//If it is smaller than 4 units, then the teamate is to close to the firing line
			if (minDist.first < 3)
			{
				if (debug)
				{
					Debug.DrawLine(teamate.position, minDist.second, Color.red);
				}
				return true;
			}
		}
		return false;

	}
}
